"""Module that allows latex output notebooks to be conditioned before
they are converted.
"""
# -----------------------------------------------------------------------------
# Copyright (c) 2013, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from traitlets import List, Unicode

from .base import Preprocessor

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


class LatexPreprocessor(Preprocessor):
    """Preprocessor for latex destined documents.

    Populates the ``latex`` key in the resources dict,
    adding definitions for pygments highlight styles.

    Sets the authors, date and title of the latex document,
    overriding the values given in the metadata.
    """

    date = Unicode(
        None,
        help=("Date of the LaTeX document"),
        allow_none=True,
    ).tag(config=True)

    title = Unicode(None, help=("Title of the LaTeX document"), allow_none=True).tag(config=True)

    author_names = List(
        Unicode(),
        default_value=None,
        help=("Author names to list in the LaTeX document"),
        allow_none=True,
    ).tag(config=True)

    style = Unicode("default", help="Name of the pygments style to use").tag(config=True)

    def preprocess(self, nb, resources):
        """Preprocessing to apply on each notebook.

        Parameters
        ----------
        nb : NotebookNode
            Notebook being converted
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        """
        # Generate Pygments definitions for Latex
        from pygments.formatters import LatexFormatter  # noqa: PLC0415

        resources.setdefault("latex", {})
        resources["latex"].setdefault(
            "pygments_definitions", LatexFormatter(style=self.style).get_style_defs()
        )
        resources["latex"].setdefault("pygments_style_name", self.style)

        if self.author_names is not None:
            nb.metadata["authors"] = [{"name": author} for author in self.author_names]

        if self.date is not None:
            nb.metadata["date"] = self.date

        if self.title is not None:
            nb.metadata["title"] = self.title

        return nb, resources



"""Export to PDF via a headless browser"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import asyncio
import concurrent.futures
import os
import subprocess
import sys
import tempfile
from importlib import util as importlib_util

from traitlets import Bool, Int, List, Unicode, default

from .html import HTMLExporter

PLAYWRIGHT_INSTALLED = importlib_util.find_spec("playwright") is not None
IS_WINDOWS = os.name == "nt"


class WebPDFExporter(HTMLExporter):
    """Writer designed to write to PDF files.

    This inherits from :class:`HTMLExporter`. It creates the HTML using the
    template machinery, and then run playwright to create a pdf.
    """

    export_from_notebook = "PDF via HTML"

    allow_chromium_download = Bool(
        False,
        help="Whether to allow downloading Chromium if no suitable version is found on the system.",
    ).tag(config=True)

    paginate = Bool(
        True,
        help="""
        Split generated notebook into multiple pages.

        If False, a PDF with one long page will be generated.

        Set to True to match behavior of LaTeX based PDF generator
        """,
    ).tag(config=True)

    page_render_timeout = Int(
        100,
        help="""
        Time to wait for the page to render before converting to PDF, in milliseconds.
        Increase this value if your notebook has a lot of complex JavaScript
        output that needs more time to load.
        """,
    ).tag(config=True)

    @default("file_extension")
    def _file_extension_default(self):
        return ".pdf"

    @default("template_extension")
    def _template_extension_default(self):
        # NOTE: we use .html.j2 so that the HTMLExporter can find the template
        return ".html.j2"

    @default("template_name")
    def _template_name_default(self):
        return "webpdf"

    disable_sandbox = Bool(
        False,
        help="""
        Disable chromium security sandbox when converting to PDF.

        WARNING: This could cause arbitrary code execution in specific circumstances,
        where JS in your notebook can execute serverside code! Please use with
        caution.

        ``https://github.com/puppeteer/puppeteer/blob/main@%7B2020-12-14T17:22:24Z%7D/docs/troubleshooting.md#setting-up-chrome-linux-sandbox``
        has more information.

        This is required for webpdf to work inside most container environments.
        """,
    ).tag(config=True)

    browser_args = List(
        Unicode(),
        help="""
        Additional arguments to pass to the browser rendering to PDF.

        These arguments will be passed directly to the browser launch method
        and can be used to customize browser behavior beyond the default settings.
        """,
    ).tag(config=True)

    def run_playwright(self, html):
        """Run playwright."""

        async def main(temp_file):
            """Run main playwright script."""

            try:
                from playwright.async_api import (  # type: ignore[import-not-found] # noqa: PLC0415,
                    async_playwright,
                )
            except ModuleNotFoundError as e:
                msg = (
                    "Playwright is not installed to support Web PDF conversion. "
                    "Please install `nbconvert[webpdf]` to enable."
                )
                raise RuntimeError(msg) from e

            if self.allow_chromium_download:
                cmd = [sys.executable, "-m", "playwright", "install", "chromium"]
                subprocess.check_call(cmd)  # noqa: S603

            playwright = await async_playwright().start()
            chromium = playwright.chromium

            args = self.browser_args
            if self.disable_sandbox:
                args.append("--no-sandbox")

            try:
                browser = await chromium.launch(
                    handle_sigint=False, handle_sigterm=False, handle_sighup=False, args=args
                )
            except Exception as e:
                msg = (
                    "No suitable chromium executable found on the system. "
                    "Please use '--allow-chromium-download' to allow downloading one,"
                    "or install it using `playwright install chromium`."
                )
                await playwright.stop()
                raise RuntimeError(msg) from e

            page = await browser.new_page()
            await page.emulate_media(media="print")
            await page.wait_for_timeout(100)
            await page.goto(f"file://{temp_file.name}", wait_until="networkidle")
            await page.wait_for_timeout(self.page_render_timeout)

            pdf_params = {"print_background": True}
            if not self.paginate:
                # Floating point precision errors cause the printed
                # PDF from spilling over a new page by a pixel fraction.
                dimensions = await page.evaluate(
                    """() => {
                    const rect = document.body.getBoundingClientRect();
                    return {
                    width: Math.ceil(rect.width) + 1,
                    height: Math.ceil(rect.height) + 1,
                    }
                }"""
                )
                width = dimensions["width"]
                height = dimensions["height"]
                # 200 inches is the maximum size for Adobe Acrobat Reader.
                pdf_params.update(
                    {
                        "width": min(width, 200 * 72),
                        "height": min(height, 200 * 72),
                    }
                )
            pdf_data = await page.pdf(**pdf_params)

            await browser.close()
            await playwright.stop()
            return pdf_data

        pool = concurrent.futures.ThreadPoolExecutor()
        # Create a temporary file to pass the HTML code to Chromium:
        # Unfortunately, tempfile on Windows does not allow for an already open
        # file to be opened by a separate process. So we must close it first
        # before calling Chromium. We also specify delete=False to ensure the
        # file is not deleted after closing (the default behavior).
        temp_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            suffix=".html", delete=False
        )
        with temp_file:
            temp_file.write(html.encode("utf-8"))
        try:
            pdf_data = pool.submit(asyncio.run, main(temp_file)).result()
        finally:
            # Ensure the file is deleted even if playwright raises an exception
            os.unlink(temp_file.name)
        return pdf_data

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from a notebook node."""
        html, resources = super().from_notebook_node(nb, resources=resources, **kw)

        self.log.info("Building PDF")
        pdf_data = self.run_playwright(html)
        self.log.info("PDF successfully created")

        return pdf_data, resources



"""reStructuredText Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from traitlets import default
from traitlets.config import Config

from ..filters import DataTypeFilter
from .templateexporter import TemplateExporter


class RSTExporter(TemplateExporter):
    """
    Exports reStructuredText documents.
    """

    @default("file_extension")
    def _file_extension_default(self):
        return ".rst"

    @default("template_name")
    def _template_name_default(self):
        return "rst"

    @default("raw_mimetypes")
    def _raw_mimetypes_default(self):
        # Up to summer 2024, nbconvert had a mistaken output_mimetype.
        # Listing that as an extra option here maintains compatibility for
        # notebooks with raw cells marked as that mimetype.
        return [self.output_mimetype, "text/restructuredtext", ""]

    output_mimetype = "text/x-rst"
    export_from_notebook = "reST"

    def default_filters(self):
        """Override filter_data_type to use native rst outputs"""
        dtf = DataTypeFilter()
        dtf.display_data_priority = [self.output_mimetype, *dtf.display_data_priority]
        filters = dict(super().default_filters())
        filters["filter_data_type"] = dtf
        return filters.items()

    @property
    def default_config(self):
        c = Config(
            {
                "CoalesceStreamsPreprocessor": {"enabled": True},
                "ExtractOutputPreprocessor": {"enabled": True},
                "HighlightMagicsPreprocessor": {"enabled": True},
            }
        )
        if super().default_config:
            c2 = super().default_config.copy()
            c2.merge(c)
            c = c2
        return c



"""HTML Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import base64
import json
import mimetypes
import os
from pathlib import Path
from typing import Any, Optional

import jinja2
import markupsafe
from bs4 import BeautifulSoup  # type: ignore[import-not-found]
from jupyter_core.paths import jupyter_path
from traitlets import Bool, Dict, Unicode, default, validate
from traitlets.config import Config

if tuple(int(x) for x in jinja2.__version__.split(".")[:3]) < (3, 0, 0):
    from jinja2 import contextfilter  # type:ignore[attr-defined]
else:
    from jinja2 import pass_context as contextfilter

from jinja2.loaders import split_template_path
from nbformat import NotebookNode

from nbconvert.filters.highlight import Highlight2HTML
from nbconvert.filters.markdown_mistune import IPythonRenderer, MarkdownWithMath
from nbconvert.filters.widgetsdatatypefilter import WidgetsDataTypeFilter
from nbconvert.utils.iso639_1 import iso639_1

from .templateexporter import TemplateExporter


def find_lab_theme(theme_name):
    """
    Find a JupyterLab theme location by name.

    Parameters
    ----------
    theme_name : str
        The name of the labextension theme you want to find.

    Raises
    ------
    ValueError
        If the theme was not found, or if it was not specific enough.

    Returns
    -------
    theme_name: str
        Full theme name (with scope, if any)
    labextension_path : Path
        The path to the found labextension on the system.
    """
    paths = jupyter_path("labextensions")

    matching_themes = []
    theme_path = None
    for path in paths:
        for dirpath, dirnames, filenames in os.walk(path):
            # If it's a federated labextension that contains themes
            if "package.json" in filenames and "themes" in dirnames:
                # TODO Find the theme name in the JS code instead?
                # TODO Find if it's a light or dark theme?
                with open(Path(dirpath) / "package.json", encoding="utf-8") as fobj:
                    labext_name = json.loads(fobj.read())["name"]

                if labext_name == theme_name or theme_name in labext_name.split("/"):
                    matching_themes.append(labext_name)

                    full_theme_name = labext_name
                    theme_path = Path(dirpath) / "themes" / labext_name

    if len(matching_themes) == 0:
        msg = f'Could not find lab theme "{theme_name}"'
        raise ValueError(msg)

    if len(matching_themes) > 1:
        msg = (
            f'Found multiple themes matching "{theme_name}": {matching_themes}. '
            "Please be more specific about which theme you want to use."
        )
        raise ValueError(msg)

    return full_theme_name, theme_path


class HTMLExporter(TemplateExporter):
    """
    Exports a basic HTML document.  This exporter assists with the export of
    HTML.  Inherit from it if you are writing your own HTML template and need
    custom preprocessors/filters.  If you don't need custom preprocessors/
    filters, just change the 'template_file' config option.
    """

    export_from_notebook = "HTML"

    anchor_link_text = Unicode("¶", help="The text used as the text for anchor links.").tag(
        config=True
    )

    exclude_anchor_links = Bool(False, help="If anchor links should be included or not.").tag(
        config=True
    )

    require_js_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js",
        help="""
        URL to load require.js from.

        Defaults to loading from cdnjs.
        """,
    ).tag(config=True)

    mathjax_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/latest.js?config=TeX-AMS_CHTML-full,Safe",
        help="""
        URL to load Mathjax from.

        Defaults to loading from cdnjs.
        """,
    ).tag(config=True)

    mermaid_js_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/mermaid/11.10.0/mermaid.esm.min.mjs",
        help="""
        URL to load MermaidJS from.

        Defaults to loading from cdnjs.
        """,
    )

    mermaid_layout_elk_js_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/mermaid-layout-elk/0.1.9/mermaid-layout-elk.esm.min.mjs",
        help="""
        URL to load MermaidJS ELK layout from.

        Defaults to loading from cdnjs.
        """,
    )

    jquery_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/jquery/2.0.3/jquery.min.js",
        help="""
        URL to load jQuery from.

        Defaults to loading from cdnjs.
        """,
    ).tag(config=True)

    jupyter_widgets_base_url = Unicode(
        "https://unpkg.com/", help="URL base for Jupyter widgets"
    ).tag(config=True)

    widget_renderer_url = Unicode("", help="Full URL for Jupyter widgets").tag(config=True)

    html_manager_semver_range = Unicode(
        "*", help="Semver range for Jupyter widgets HTML manager"
    ).tag(config=True)

    @default("file_extension")
    def _file_extension_default(self):
        return ".html"

    @default("template_name")
    def _template_name_default(self):
        return "lab"

    theme = Unicode(
        "light",
        help="Template specific theme(e.g. the name of a JupyterLab CSS theme distributed as prebuilt extension for the lab template)",
    ).tag(config=True)

    sanitize_html = Bool(
        False,
        help=(
            "Whether the HTML in Markdown cells and cell outputs should be sanitized."
            "This should be set to True by nbviewer or similar tools."
        ),
    ).tag(config=True)

    skip_svg_encoding = Bool(
        False,
        help=("Whether the svg to image data attribute encoding should occur"),
    ).tag(config=True)

    embed_images = Bool(
        False, help="Whether or not to embed images as base64 in markdown cells."
    ).tag(config=True)

    output_mimetype = "text/html"

    lexer_options = Dict(
        {},
        help=(
            "Options to be passed to the pygments lexer for highlighting markdown code blocks. "
            "See https://pygments.org/docs/lexers/#available-lexers for available options."
        ),
    ).tag(config=True)

    @property
    def default_config(self):
        c = Config(
            {
                "NbConvertBase": {
                    "display_data_priority": [
                        "application/vnd.jupyter.widget-view+json",
                        "application/javascript",
                        "text/html",
                        "text/markdown",
                        "image/svg+xml",
                        "text/vnd.mermaid",
                        "text/latex",
                        "image/png",
                        "image/jpeg",
                        "text/plain",
                    ]
                },
                "HighlightMagicsPreprocessor": {"enabled": True},
            }
        )
        if super().default_config:
            c2 = super().default_config.copy()
            c2.merge(c)
            c = c2
        return c

    language_code = Unicode(
        "en", help="Language code of the content, should be one of the ISO639-1"
    ).tag(config=True)

    @validate("language_code")
    def _valid_language_code(self, proposal):
        if self.language_code not in iso639_1:
            self.log.warning(
                '"%s" is not an ISO 639-1 language code. '
                'It has been replaced by the default value "en".',
                self.language_code,
            )
            return proposal["trait"].default_value
        return proposal["value"]

    @contextfilter
    def markdown2html(self, context, source):
        """Markdown to HTML filter respecting the anchor_link_text setting"""
        cell = context.get("cell", {})
        attachments = cell.get("attachments", {})
        path = context.get("resources", {}).get("metadata", {}).get("path", "")

        renderer = IPythonRenderer(
            escape=False,
            attachments=attachments,
            embed_images=self.embed_images,
            path=path,
            anchor_link_text=self.anchor_link_text,
            exclude_anchor_links=self.exclude_anchor_links,
            **self.lexer_options,
        )
        return MarkdownWithMath(renderer=renderer).render(source)

    def default_filters(self):
        """Get the default filters."""
        yield from super().default_filters()
        yield ("markdown2html", self.markdown2html)

    def from_notebook_node(  # type:ignore[override]
        self, nb: NotebookNode, resources: Optional[dict[str, Any]] = None, **kw: Any
    ) -> tuple[str, dict[str, Any]]:
        """Convert from notebook node."""
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get(
            "highlight_code", Highlight2HTML(pygments_lexer=lexer, parent=self)
        )

        resources = self._init_resources(resources)

        filter_data_type = WidgetsDataTypeFilter(
            notebook_metadata=self._nb_metadata, parent=self, resources=resources
        )

        self.register_filter("highlight_code", highlight_code)
        self.register_filter("filter_data_type", filter_data_type)
        html, resources = super().from_notebook_node(nb, resources, **kw)
        soup = BeautifulSoup(html, features="html.parser")
        # Add image's alternative text
        missing_alt = 0
        for elem in soup.select("img:not([alt])"):
            elem.attrs["alt"] = "No description has been provided for this image"
            missing_alt += 1
        if missing_alt:
            self.log.warning("Alternative text is missing on %s image(s).", missing_alt)
        # Set input and output focusable
        for elem in soup.select(".jp-Notebook div.jp-Cell-inputWrapper"):
            elem.attrs["tabindex"] = "0"
        for elem in soup.select(".jp-Notebook div.jp-OutputArea-output"):
            elem.attrs["tabindex"] = "0"

        return str(soup), resources

    def _init_resources(self, resources):
        def resources_include_css(name):
            env = self.environment
            code = """<style type="text/css">\n%s</style>""" % (env.loader.get_source(env, name)[0])
            return markupsafe.Markup(code)  # noqa:S704

        def resources_include_lab_theme(name):
            # Try to find the theme with the given name, looking through the labextensions
            _, theme_path = find_lab_theme(name)

            with open(theme_path / "index.css") as file:
                data = file.read()

            # Embed assets (fonts, images...)
            for asset in os.listdir(theme_path):
                local_url = f"url({Path(asset).as_posix()})"

                if local_url in data:
                    mime_type = mimetypes.guess_type(asset)[0]

                    # Replace asset url by a base64 dataurl
                    with open(theme_path / asset, "rb") as assetfile:
                        base64_data = base64.b64encode(assetfile.read())
                        base64_str = base64_data.replace(b"\n", b"").decode("ascii")

                        data = data.replace(local_url, f"url(data:{mime_type};base64,{base64_str})")

            code = """<style type="text/css">\n%s</style>""" % data
            return markupsafe.Markup(code)  # noqa:S704

        def resources_include_js(name, module=False):
            """Get the resources include JS for a name. If module=True, import as ES module"""
            env = self.environment
            code = f"""<script {'type="module"' if module else ""}>\n{env.loader.get_source(env, name)[0]}</script>"""
            return markupsafe.Markup(code)  # noqa:S704

        def resources_include_url(name):
            """Get the resources include url for a name."""
            env = self.environment
            mime_type, _encoding = mimetypes.guess_type(name)
            try:
                # we try to load via the jinja loader, but that tries to load
                # as (encoded) text
                data = env.loader.get_source(env, name)[0].encode("utf8")
            except UnicodeDecodeError:
                # if that fails (for instance a binary file, png or ttf)
                # we mimic jinja2
                pieces = split_template_path(name)
                for searchpath in self.template_paths:
                    filename = os.path.join(searchpath, *pieces)
                    if os.path.exists(filename):
                        with open(filename, "rb") as f:
                            data = f.read()
                            break
                else:
                    msg = f"No file {name!r} found in {searchpath!r}"
                    raise ValueError(msg)
            data = base64.b64encode(data)
            data = data.replace(b"\n", b"").decode("ascii")
            src = f"data:{mime_type};base64,{data}"
            return markupsafe.Markup(src)  # noqa:S704

        resources = super()._init_resources(resources)
        resources["theme"] = self.theme
        resources["include_css"] = resources_include_css
        resources["include_lab_theme"] = resources_include_lab_theme
        resources["include_js"] = resources_include_js
        resources["include_url"] = resources_include_url
        resources["require_js_url"] = self.require_js_url
        resources["mathjax_url"] = self.mathjax_url
        resources["mermaid_js_url"] = self.mermaid_js_url
        resources["mermaid_layout_elk_js_url"] = self.mermaid_layout_elk_js_url
        resources["jquery_url"] = self.jquery_url
        resources["jupyter_widgets_base_url"] = self.jupyter_widgets_base_url
        resources["widget_renderer_url"] = self.widget_renderer_url
        resources["html_manager_semver_range"] = self.html_manager_semver_range
        resources["should_sanitize_html"] = self.sanitize_html
        resources["language_code"] = self.language_code
        resources["should_not_encode_svg"] = self.skip_svg_encoding
        return resources



"""PostProcessor for serving reveal.js HTML slideshows."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import os
import threading
import typing as t
import webbrowser

from tornado import gen, httpserver, ioloop, log, web
from tornado.httpclient import AsyncHTTPClient
from traitlets import Bool, Int, Unicode

from .base import PostProcessorBase


class ProxyHandler(web.RequestHandler):
    """handler the proxies requests from a local prefix to a CDN"""

    @gen.coroutine
    def get(self, prefix, url):
        """proxy a request to a CDN"""
        proxy_url = "/".join([self.settings["cdn"], url])
        client = self.settings["client"]
        response = yield client.fetch(proxy_url)

        for header in ["Content-Type", "Cache-Control", "Date", "Last-Modified", "Expires"]:
            if header in response.headers:
                self.set_header(header, response.headers[header])
        self.finish(response.body)


class ServePostProcessor(PostProcessorBase):
    """Post processor designed to serve files

    Proxies reveal.js requests to a CDN if no local reveal.js is present
    """

    open_in_browser = Bool(True, help="""Should the browser be opened automatically?""").tag(
        config=True
    )

    browser = Unicode(
        "",
        help="""Specify what browser should be used to open slides. See
                      https://docs.python.org/3/library/webbrowser.html#webbrowser.register
                      to see how keys are mapped to browser executables. If
                      not specified, the default browser will be determined
                      by the `webbrowser`
                      standard library module, which allows setting of the BROWSER
                      environment variable to override it.
                      """,
    ).tag(config=True)

    reveal_cdn = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.5.0", help="""URL for reveal.js CDN."""
    ).tag(config=True)
    reveal_prefix = Unicode("reveal.js", help="URL prefix for reveal.js").tag(config=True)
    ip = Unicode("127.0.0.1", help="The IP address to listen on.").tag(config=True)
    port = Int(8000, help="port for the server to listen on.").tag(config=True)

    def postprocess(self, input):
        """Serve the build directory with a webserver."""
        dirname, filename = os.path.split(input)
        handlers: list[tuple[t.Any, ...]] = [
            (r"/(.+)", web.StaticFileHandler, {"path": dirname}),
            (r"/", web.RedirectHandler, {"url": "/%s" % filename}),
        ]

        if "://" in self.reveal_prefix or self.reveal_prefix.startswith("//"):
            # reveal specifically from CDN, nothing to do
            pass
        elif os.path.isdir(os.path.join(dirname, self.reveal_prefix)):
            # reveal prefix exists
            self.log.info("Serving local %s", self.reveal_prefix)
        else:
            self.log.info("Redirecting %s requests to %s", self.reveal_prefix, self.reveal_cdn)
            handlers.insert(0, (r"/(%s)/(.*)" % self.reveal_prefix, ProxyHandler))

        app = web.Application(
            handlers,
            cdn=self.reveal_cdn,
            client=AsyncHTTPClient(),
        )

        # hook up tornado logging to our logger
        log.app_log = self.log

        http_server = httpserver.HTTPServer(app)
        http_server.listen(self.port, address=self.ip)
        url = "http://%s:%i/%s" % (self.ip, self.port, filename)
        print("Serving your slides at %s" % url)
        print("Use Control-C to stop this server")
        if self.open_in_browser:
            try:
                browser = webbrowser.get(self.browser or None)
                b = lambda: browser.open(url, new=2)  # noqa: E731
                threading.Thread(target=b).start()
            except webbrowser.Error as e:
                self.log.warning("No web browser found: %s.", e)
                browser = None

        try:
            ioloop.IOLoop.instance().start()
        except KeyboardInterrupt:
            print("\nInterrupted")


def main(path):
    """allow running this module to serve the slides"""
    server = ServePostProcessor()
    server(path)


if __name__ == "__main__":
    import sys

    main(sys.argv[1])



from .asciidoc import ASCIIDocExporter
from .base import ExporterDisabledError, ExporterNameError, export, get_export_names, get_exporter
from .exporter import Exporter, FilenameExtension, ResourcesDict
from .html import HTMLExporter
from .latex import LatexExporter
from .markdown import MarkdownExporter
from .notebook import NotebookExporter
from .pdf import PDFExporter
from .python import PythonExporter
from .qtpdf import QtPDFExporter
from .qtpng import QtPNGExporter
from .rst import RSTExporter
from .script import ScriptExporter
from .slides import SlidesExporter
from .templateexporter import TemplateExporter
from .webpdf import WebPDFExporter

__all__ = [
    "ASCIIDocExporter",
    "Exporter",
    "ExporterDisabledError",
    "ExporterNameError",
    "FilenameExtension",
    "HTMLExporter",
    "LatexExporter",
    "MarkdownExporter",
    "NotebookExporter",
    "PDFExporter",
    "PythonExporter",
    "QtPDFExporter",
    "QtPNGExporter",
    "RSTExporter",
    "ResourcesDict",
    "ScriptExporter",
    "SlidesExporter",
    "TemplateExporter",
    "WebPDFExporter",
    "export",
    "get_export_names",
    "get_exporter",
]



"""A qt exporter."""

import os
import sys
import tempfile
import time

from traitlets import default

from .html import HTMLExporter


class QtExporter(HTMLExporter):
    """A qt exporter."""

    paginate = None
    format = ""

    @default("file_extension")
    def _file_extension_default(self):
        return ".html"

    def _check_launch_reqs(self):
        if sys.platform.startswith("win") and self.format == "png":
            msg = "Exporting to PNG using Qt is currently not supported on Windows."
            raise RuntimeError(msg)
        from .qt_screenshot import QT_INSTALLED  # noqa: PLC0415

        if not QT_INSTALLED:
            msg = (
                f"PyQtWebEngine is not installed to support Qt {self.format.upper()} conversion. "
                f"Please install `nbconvert[qt{self.format}]` to enable."
            )
            raise RuntimeError(msg)
        from .qt_screenshot import QtScreenshot  # noqa: PLC0415

        return QtScreenshot

    def _run_pyqtwebengine(self, html):
        ext = ".html"
        temp_file = tempfile.NamedTemporaryFile(  # noqa: SIM115
            suffix=ext, delete=False
        )
        filename = f"{temp_file.name[: -len(ext)]}.{self.format}"
        with temp_file:
            temp_file.write(html.encode("utf-8"))
        try:
            QtScreenshot = self._check_launch_reqs()
            s = QtScreenshot()
            s.capture(f"file://{temp_file.name}", filename, self.paginate)
        finally:
            # Ensure the file is deleted even if pyqtwebengine raises an exception
            os.unlink(temp_file.name)
        # Prefer Qt's in-memory bytes, but fall back to reading the file on disk
        data = getattr(s, "data", b"")

        if (not data) and os.path.exists(filename):
            deadline = time.time() + 5.0
            while time.time() < deadline:
                try:
                    if os.path.getsize(filename) > 0:
                        break
                except OSError:
                    pass
                time.sleep(0.05)

            if os.path.exists(filename) and os.path.getsize(filename) > 0:
                with open(filename, "rb") as f:
                    data = f.read()

        # Best-effort cleanup of the generated output file
        try:
            if os.path.exists(filename):
                os.unlink(filename)
        except OSError:
            pass

        return data

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from notebook node."""
        self._check_launch_reqs()
        html, resources = super().from_notebook_node(nb, resources=resources, **kw)

        self.log.info("Building %s", self.format.upper())
        data = self._run_pyqtwebengine(html)
        self.log.info("%s successfully created", self.format.upper())

        # convert output extension
        # the writer above required it to be html
        resources["output_extension"] = f".{self.format}"

        return data, resources



"""
Module containing a preprocessor that removes cells if they match
one or more regular expression.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

from traitlets import Set, Unicode

from .base import Preprocessor


class TagRemovePreprocessor(Preprocessor):
    """
    Removes inputs, outputs, or cells from a notebook that
    have tags that designate they are to be removed prior to exporting
    the notebook.

    remove_cell_tags
        removes cells tagged with these values

    remove_all_outputs_tags
        removes entire output areas on cells
        tagged with these values

    remove_single_output_tags
        removes individual output objects on
        outputs tagged with these values

    remove_input_tags
        removes inputs tagged with these values
    """

    remove_cell_tags: set[str] = Set(  # type:ignore[assignment]
        Unicode(),
        default_value=[],
        help=(
            "Tags indicating which cells are to be removed,matches tags in ``cell.metadata.tags``."
        ),
    ).tag(config=True)
    remove_all_outputs_tags: set[str] = Set(  # type:ignore[assignment]
        Unicode(),
        default_value=[],
        help=(
            "Tags indicating cells for which the outputs are to be removed,"
            "matches tags in ``cell.metadata.tags``."
        ),
    ).tag(config=True)
    remove_single_output_tags: set[str] = Set(  # type:ignore[assignment]
        Unicode(),
        default_value=[],
        help=(
            "Tags indicating which individual outputs are to be removed,"
            "matches output *i* tags in ``cell.outputs[i].metadata.tags``."
        ),
    ).tag(config=True)
    remove_input_tags: set[str] = Set(  # type:ignore[assignment]
        Unicode(),
        default_value=[],
        help=(
            "Tags indicating cells for which input is to be removed,"
            "matches tags in ``cell.metadata.tags``."
        ),
    ).tag(config=True)
    remove_metadata_fields: set[str] = Set({"collapsed", "scrolled"}).tag(config=True)  # type:ignore[assignment]

    def check_cell_conditions(self, cell, resources, index):
        """
        Checks that a cell has a tag that is to be removed

        Returns: Boolean.
        True means cell should *not* be removed.
        """

        # Return true if any of the tags in the cell are removable.
        return not self.remove_cell_tags.intersection(cell.get("metadata", {}).get("tags", []))

    def preprocess(self, nb, resources):
        """
        Preprocessing to apply to each notebook. See base.py for details.
        """
        # Skip preprocessing if the list of patterns is empty
        if not any(
            [
                self.remove_cell_tags,
                self.remove_all_outputs_tags,
                self.remove_single_output_tags,
                self.remove_input_tags,
            ]
        ):
            return nb, resources

        # Filter out cells that meet the conditions
        nb.cells = [
            self.preprocess_cell(cell, resources, index)[0]
            for index, cell in enumerate(nb.cells)
            if self.check_cell_conditions(cell, resources, index)
        ]

        return nb, resources

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Apply a transformation on each cell. See base.py for details.
        """

        if (
            self.remove_all_outputs_tags.intersection(cell.get("metadata", {}).get("tags", []))
            and cell.cell_type == "code"
        ):
            cell.outputs = []
            cell.execution_count = None
            # Remove metadata associated with output
            if "metadata" in cell:
                for field in self.remove_metadata_fields:
                    cell.metadata.pop(field, None)

        if self.remove_input_tags.intersection(cell.get("metadata", {}).get("tags", [])):
            cell.metadata["transient"] = {"remove_source": True}

        if cell.get("outputs", []):
            cell.outputs = [
                output
                for output_index, output in enumerate(cell.outputs)
                if self.check_output_conditions(output, resources, cell_index, output_index)
            ]
        return cell, resources

    def check_output_conditions(self, output, resources, cell_index, output_index):
        """
        Checks that an output has a tag that indicates removal.

        Returns: Boolean.
        True means output should *not* be removed.
        """
        return not self.remove_single_output_tags.intersection(
            output.get("metadata", {}).get("tags", [])
        )



"""String filters.

Contains a collection of useful string manipulation filters for use in Jinja
templates.
"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import base64
import os
import re
import textwrap
import warnings
from urllib.parse import quote
from xml.etree.ElementTree import Element

import bleach

# defusedxml does safe(r) parsing of untrusted XML data
from defusedxml import ElementTree

from nbconvert.preprocessors.sanitize import _get_default_css_sanitizer

__all__ = [
    "add_anchor",
    "add_prompts",
    "ascii_only",
    "clean_html",
    "comment_lines",
    "get_lines",
    "html2text",
    "ipython2python",
    "path2url",
    "posix_path",
    "prevent_list_blocks",
    "strip_dollars",
    "strip_files_prefix",
    "strip_trailing_newline",
    "text_base64",
    "wrap_text",
]


def wrap_text(text, width=100):
    """
    Intelligently wrap text.
    Wrap text without breaking words if possible.

    Parameters
    ----------
    text : str
        Text to wrap.
    width : int, optional
        Number of characters to wrap to, default 100.
    """

    split_text = text.split("\n")
    wrp = map(lambda x: textwrap.wrap(x, width), split_text)  # noqa: C417
    wrpd = map("\n".join, wrp)
    return "\n".join(wrpd)


def html2text(element):
    """extract inner text from html

    Analog of jQuery's $(element).text()
    """
    if isinstance(element, (str,)):
        try:
            element = ElementTree.fromstring(element)
        except Exception:
            # failed to parse, just return it unmodified
            return element

    text = element.text or ""
    for child in element:
        text += html2text(child)
    text += element.tail or ""
    return text


def clean_html(element):
    """Clean an html element."""
    element = element.decode() if isinstance(element, bytes) else str(element)
    kwargs = {}
    css_sanitizer = _get_default_css_sanitizer()
    if css_sanitizer:
        kwargs["css_sanitizer"] = css_sanitizer
    return bleach.clean(
        element,
        tags=[*bleach.ALLOWED_TAGS, "div", "pre", "code", "span", "table", "tr", "td"],
        attributes={
            **bleach.ALLOWED_ATTRIBUTES,
            "*": ["class", "id"],
        },
        **kwargs,
    )


def _convert_header_id(header_contents):
    """Convert header contents to valid id value. Takes string as input, returns string.

    Note: this may be subject to change in the case of changes to how we wish to generate ids.

    For use on markdown headings.
    """
    # Valid IDs need to be non-empty and contain no space characters, but are otherwise arbitrary.
    # However, these IDs are also used in URL fragments, which are more restrictive, so we URL
    # encode any characters that are not valid in URL fragments.
    return quote(header_contents.replace(" ", "-"), safe="?/:@!$&'()*+,;=")


def add_anchor(html, anchor_link_text="¶"):
    """Add an id and an anchor-link to an html header

    For use on markdown headings
    """
    try:
        h = ElementTree.fromstring(html)
    except Exception:
        # failed to parse, just return it unmodified
        return html
    link = _convert_header_id(html2text(h))
    h.set("id", link)
    a = Element("a", {"class": "anchor-link", "href": "#" + link})
    try:
        # Test if the anchor link text is HTML (e.g. an image)
        a.append(ElementTree.fromstring(anchor_link_text))
    except Exception:
        # If we fail to parse, assume we've just got regular text
        a.text = anchor_link_text
    h.append(a)

    return ElementTree.tostring(h).decode(encoding="utf-8")


def add_prompts(code, first=">>> ", cont="... "):
    """Add prompts to code snippets"""
    new_code = []
    code_list = code.split("\n")
    new_code.append(first + code_list[0])
    for line in code_list[1:]:
        new_code.append(cont + line)
    return "\n".join(new_code)


def strip_dollars(text):
    """
    Remove all dollar symbols from text

    Parameters
    ----------
    text : str
        Text to remove dollars from
    """

    return text.strip("$")


files_url_pattern = re.compile(r'(src|href)\=([\'"]?)/?files/')
markdown_url_pattern = re.compile(r"(!?)\[(?P<caption>.*?)\]\(/?files/(?P<location>.*?)\)")


def strip_files_prefix(text):
    """
    Fix all fake URLs that start with ``files/``, stripping out the ``files/`` prefix.
    Applies to both urls (for html) and relative paths (for markdown paths).

    Parameters
    ----------
    text : str
        Text in which to replace 'src="files/real...' with 'src="real...'
    """
    cleaned_text = files_url_pattern.sub(r"\1=\2", text)
    cleaned_text = markdown_url_pattern.sub(r"\1[\2](\3)", cleaned_text)
    return cleaned_text  # noqa: RET504


def comment_lines(text, prefix="# "):
    """
    Build a Python comment line from input text.

    Parameters
    ----------
    text : str
        Text to comment out.
    prefix : str
        Character to append to the start of each line.
    """

    # Replace line breaks with line breaks and comment symbols.
    # Also add a comment symbol at the beginning to comment out
    # the first line.
    return prefix + ("\n" + prefix).join(text.split("\n"))


def get_lines(text, start=None, end=None):
    """
    Split the input text into separate lines and then return the
    lines that the caller is interested in.

    Parameters
    ----------
    text : str
        Text to parse lines from.
    start : int, optional
        First line to grab from.
    end : int, optional
        Last line to grab from.
    """

    # Split the input into lines.
    lines = text.split("\n")

    # Return the right lines.
    return "\n".join(lines[start:end])  # re-join


def ipython2python(code):
    """Transform IPython syntax to pure Python syntax

    Parameters
    ----------
    code : str
        IPython code, to be transformed to pure Python
    """
    try:
        from IPython.core.inputtransformer2 import TransformerManager  # noqa: PLC0415
    except ImportError:
        warnings.warn(
            "IPython is needed to transform IPython syntax to pure Python."
            " Install ipython if you need this functionality.",
            stacklevel=2,
        )
        return code
    else:
        isp = TransformerManager()
        return isp.transform_cell(code)


def posix_path(path):
    """Turn a path into posix-style path/to/etc

    Mainly for use in latex on Windows,
    where native Windows paths are not allowed.
    """
    if os.path.sep != "/":
        return path.replace(os.path.sep, "/")
    return path


def path2url(path):
    """Turn a file path into a URL"""
    parts = path.split(os.path.sep)
    return "/".join(quote(part) for part in parts)


def ascii_only(s):
    """ensure a string is ascii"""
    return s.encode("ascii", "replace").decode("ascii")


def prevent_list_blocks(s):
    """
    Prevent presence of enumerate or itemize blocks in latex headings cells
    """
    out = re.sub(r"(^\s*\d*)\.", r"\1\.", s)
    out = re.sub(r"(^\s*)\-", r"\1\-", out)
    out = re.sub(r"(^\s*)\+", r"\1\+", out)
    out = re.sub(r"(^\s*)\*", r"\1\*", out)
    return out  # noqa: RET504


def strip_trailing_newline(text):
    """
    Strips a newline from the end of text.
    """
    if text.endswith("\n"):
        text = text[:-1]
    return text


def text_base64(text):
    """
    Encode base64 text
    """
    return base64.b64encode(text.encode()).decode()



"""Utilities for converting notebooks to and from different formats."""

from ._version import __version__, version_info

try:
    from . import filters, postprocessors, preprocessors, writers
    from .exporters import (
        ASCIIDocExporter,
        Exporter,
        ExporterNameError,
        FilenameExtension,
        HTMLExporter,
        LatexExporter,
        MarkdownExporter,
        NotebookExporter,
        PDFExporter,
        PythonExporter,
        QtPDFExporter,
        QtPNGExporter,
        RSTExporter,
        ScriptExporter,
        SlidesExporter,
        TemplateExporter,
        WebPDFExporter,
        export,
        get_export_names,
        get_exporter,
    )
except ModuleNotFoundError:
    # We hit this condition when the package is not yet fully installed.
    pass


__all__ = [
    "ASCIIDocExporter",
    "Exporter",
    "ExporterNameError",
    "FilenameExtension",
    "HTMLExporter",
    "LatexExporter",
    "MarkdownExporter",
    "NotebookExporter",
    "PDFExporter",
    "PythonExporter",
    "QtPDFExporter",
    "QtPNGExporter",
    "RSTExporter",
    "ScriptExporter",
    "SlidesExporter",
    "TemplateExporter",
    "WebPDFExporter",
    "__version__",
    "export",
    "filters",
    "get_export_names",
    "get_exporter",
    "postprocessors",
    "preprocessors",
    "version_info",
    "writers",
]



"""Backport of Python 3.11's contextlib.chdir."""

import os
from contextlib import AbstractContextManager


class chdir(AbstractContextManager):  # type:ignore[type-arg]
    """Non thread-safe context manager to change the current working directory."""

    def __init__(self, path):
        """Initialize the manager."""
        self.path = path
        self._old_cwd = []

    def __enter__(self):
        """Enter the context."""
        self._old_cwd.append(os.getcwd())
        os.chdir(self.path)

    def __exit__(self, *excinfo):
        """Exit the context."""
        os.chdir(self._old_cwd.pop())



# Class base Preprocessors
# Backwards compatibility for imported name
from nbclient.exceptions import CellExecutionError

from .base import Preprocessor
from .clearmetadata import ClearMetadataPreprocessor
from .clearoutput import ClearOutputPreprocessor
from .coalescestreams import CoalesceStreamsPreprocessor
from .convertfigures import ConvertFiguresPreprocessor
from .csshtmlheader import CSSHTMLHeaderPreprocessor
from .execute import ExecutePreprocessor
from .extractattachments import ExtractAttachmentsPreprocessor
from .extractoutput import ExtractOutputPreprocessor
from .highlightmagics import HighlightMagicsPreprocessor
from .latex import LatexPreprocessor
from .regexremove import RegexRemovePreprocessor
from .svg2pdf import SVG2PDFPreprocessor
from .tagremove import TagRemovePreprocessor

__all__ = [
    "CSSHTMLHeaderPreprocessor",
    "CellExecutionError",
    "ClearMetadataPreprocessor",
    "ClearOutputPreprocessor",
    "CoalesceStreamsPreprocessor",
    "ConvertFiguresPreprocessor",
    "ExecutePreprocessor",
    "ExtractAttachmentsPreprocessor",
    "ExtractOutputPreprocessor",
    "HighlightMagicsPreprocessor",
    "LatexPreprocessor",
    "Preprocessor",
    "RegexRemovePreprocessor",
    "SVG2PDFPreprocessor",
    "TagRemovePreprocessor",
]



"""Export to PNG via a headless browser"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from .qt_exporter import QtExporter


class QtPNGExporter(QtExporter):
    """Writer designed to write to PNG files.

    This inherits from :class:`HTMLExporter`. It creates the HTML using the
    template machinery, and then uses pyqtwebengine to create a png.
    """

    export_from_notebook = "PNG via HTML"
    format = "png"
